# -*- coding: utf-8 -*-

from .basedataset import BaseDataset
from .benchmarkdataset import BenchmarkDataset
from .betadataset import BETADataset
from .eldbetadataset import ELDBETADataset
from .nakanishidataset import NakanishiDataset
from .openbmidataset import openBMIDataset
from .wearabledataset import WearableDataset_dry, WearableDataset_wet



# -*- coding: utf-8 -*-

import os
import numpy as np
import zipfile
import shutil

from typing import Union, Optional, Dict, List, Tuple
from numpy import ndarray, transpose

from .basedataset import BaseDataset
from .subjectinfo import SubInfo
from ..utils.download import download_single_file
from ..utils.io import loadmat

class NakanishiDataset(BaseDataset):
    """
    Nakanishi 2015

    Each .mat file has a four-way tensor electroencephalogram (EEG) data for each subject. 
    Please see the reference paper for the detail.

    size(eeg) = [Num. of targets, Num. of channels, Num. of sampling points, Num. of trials]
    Num. of Targets 	: 12
    Num. of Channels 	: 8
    Num. of sampling points : 1114
    Num. of trials 		: 15
    Sampling rate 		: 256 Hz
    * The order of the stimulus frequencies in the EEG data: 
    [9.25, 11.25, 13.25, 9.75, 11.75, 13.75, 10.25, 12.25, 14.25, 10.75, 12.75, 14.75] Hz
    (e.g., eeg(1,:,:,:) and eeg(5,:,:,:) are the EEG data while a subject was gazing at the visual stimuli flickering at 9.25 Hz and 11.75Hz, respectively.)
    * The onset of visual stimulation is at 39th sample point.

    Reference:
    Masaki Nakanishi, Yijun Wang, Yu-Te Wang and Tzyy-Ping Jung,
    "A Comparison Study of Canonical Correlation Analysis Based Methods for Detecting Steady-State Visual Evoked Potentials,"
    PLoS One, vol.10, no.10, e140703, 2015.
    """

    _CHANNELS = [
        'PO7','PO3','POz','PO4','PO8',
        'O1','Oz','O2'
    ]

    _FREQS = [
        9.25, 11.25, 13.25, 
        9.75, 11.75, 13.75, 
        10.25, 12.25, 14.25, 
        10.75, 12.75, 14.75
    ]

    _PHASES = [
        0, 0, 0,
        0.5, 0.5, 0.5,
        1, 1, 1,
        1.5, 1.5, 1.5
    ]

    _SUBJECTS = [SubInfo(ID = 's{:d}'.format(sub_idx)) for sub_idx in range(1,10+1,1)]

    def __init__(self, 
                 path: Optional[str] = None,
                 path_support_file: Optional[str] = None):
        super().__init__(subjects = self._SUBJECTS, 
                         ID = 'Nakanishi2015', 
                         url = 'ftp://sccn.ucsd.edu/pub/cca_ssvep.zip', 
                         paths = path, 
                         channels = self._CHANNELS, 
                         srate = 256, 
                         block_num = 15, 
                         trial_num = len(self._FREQS),
                         trial_len = 4, 
                         stim_info = {'stim_num': len(self._FREQS),
                                      'freqs': self._FREQS,
                                      'phases': [i * np.pi for i in self._PHASES]},
                         support_files = [],
                         path_support_file = path_support_file,
                         t_prestim = 39/256,
                         t_break = 1,
                         default_t_latency = 0.135)
        self.trial_label_check_list = {}
        for trial_i in range(self.trial_num):
            self.trial_label_check_list[trial_i] = trial_i

    def download_single_subject(self,
                                subject: SubInfo):
        source_url = self.url
        desertation = os.path.join(subject.path, 'cca_ssvep.zip')
        
        data_file = os.path.join(subject.path, subject.ID + '.mat')

        download_flag = True
        
        if not os.path.isfile(data_file):
            try:
                download_single_file(source_url, desertation, progressbar = False)
            
                with zipfile.ZipFile(desertation,'r') as archive:
                    archive.extractall(subject.path)
                    
                os.remove(desertation)

                for (dirpath, dirnames, filenames) in os.walk(os.path.join(subject.path, 'cca_ssvep')):
                    for filename in filenames:
                        src = os.path.join(dirpath, filename)
                        dst = os.path.join(subject.path, filename)
                        shutil.copyfile(src, dst)
                shutil.rmtree(os.path.join(subject.path, 'cca_ssvep'))
            except:
                download_flag = False

        return download_flag, source_url, desertation

    def download_file(self,
                      file_name: str):
        return True, None, None

    def get_sub_data(self, 
                     sub_idx: int) -> ndarray:
        if sub_idx < 0:
            raise ValueError('Subject index cannot be negative')
        if sub_idx > len(self.subjects)-1:
            raise ValueError('Subject index should be smaller than {:d}'.format(len(self.subjects)))
        
        sub_info = self.subjects[sub_idx]
        file_path = os.path.join(sub_info.path, sub_info.ID + '.mat')
        
        mat_data = loadmat(file_path)
        data = mat_data['eeg']
        data = transpose(data, (3,0,1,2)) # block_num * stimulus_num * ch_num * whole_trial_samples

        return data

    def get_label_single_trial(self,
                               sub_idx: int,
                               block_idx: int,
                               trial_idx: int) -> int:
        return self.trial_label_check_list[trial_idx]


# -*- coding: utf-8 -*-

from typing import Union, Optional, Dict, List, Tuple
from numpy import ndarray, linspace, pi, sin, cos, expand_dims, concatenate
import numpy as np
from scipy import signal

def sine_snr(X : ndarray,
             ref : ndarray):
    """
    Calculate SNR using reference
    """
    ref = ref/ref.shape[1]
    ref_mul = ref.T @ ref
    X_ref = np.trace(X @ ref.T @ ref @ X.T)
    X_non_ref = np.trace(X @ (np.eye(ref.shape[1]) - ref_mul) @ X.T)
    return 10*np.log10(X_ref/X_non_ref)

def freqs_snr(X : ndarray,
              target_fre : float,
              srate : float,
              Nh : int,
              detrend_flag : bool = True,
              NFFT : Optional[int] = None):
    """
    Calculate FFT and then calculate SNR
    """
    freq, fft_res = fft(X, srate, detrend_flag = detrend_flag, NFFT = NFFT)
    abs_fft_res = np.abs(fft_res)

    stim_amp = 0
    for n in range(Nh):
        freLoc = np.argmin(np.abs(freq - (target_fre*(n+1))))
        stim_amp += abs_fft_res[0,freLoc]
    snr = 10*np.log10(stim_amp/(np.sum(abs_fft_res)-stim_amp))
    return snr

def freqs_phase(X : ndarray,
                target_fre : float,
                target_phase : float,
                srate : float,
                detrend_flag : bool = True,
                NFFT : Optional[int] = None):
    """
    Calculate FFT and then calculate phase
    """
    freq, fft_res = fft(X, srate, detrend_flag = detrend_flag, NFFT = NFFT)
    angle_fft_res = np.angle(fft_res)
    freLoc = np.argmin(np.abs(freq - target_fre))
    stim_phase = angle_fft_res[0,freLoc]
    if stim_phase != target_phase:
        k1 = np.floor((target_phase - stim_phase)/(2*np.pi))
        k2 = -k1
        k3 = np.floor((stim_phase - target_phase)/(2*np.pi))
        k4 = -k3
        k = np.array([k1,k2,k3,k4])
        k_loc = np.argmin(np.abs(stim_phase + 2*np.pi*k - target_phase))
        stim_phase = stim_phase + 2*np.pi*k[k_loc]
    return stim_phase


def nextpow2(n):
    '''
    Retrun the first P such that 2 ** P >= abs(n).
    '''
    return np.ceil(np.log2(np.abs(n)))

def fft(X : ndarray,
        fs : float,
        detrend_flag : bool = True,
        NFFT : Optional[int] = None):
    """
    Calculate FFT

    Parameters
    -----------
    X : ndarray
        Input signal. The shape is (1*N) where N is the sampling number.
    fs : float
        Sampling freqeuncy.
    detrend_flag : bool
        Whether detrend. If True, X will be detrended first. Default is True.
    NFFT : Optional[int]
        Number of FFT. If None, NFFT is equal to 2^nextpow2(X.shape[1]). Default is None.

    Returns
    -------------
    freqs : ndarray
        Corresponding frequencies
    fft_res : ndarray
        FFT result
    """
    X_raw, X_col = X.shape
    if X_raw!=1:
        raise ValueError('The row number of the input signal for the FFT must be 1.')
    if X_col==1:
        raise ValueError('The column number of the input signal for the FFT cannot be 1.')
    if NFFT is None:
        NFFT = 2 ** nextpow2(X_col)
    if type(NFFT) is not int:
        NFFT = int(NFFT)
    
    if detrend_flag:
        X = signal.detrend(X, axis = 1)

    fft_res = np.fft.fft(X, NFFT, axis = 1)
    freqs = np.fft.fftfreq(NFFT, 1/fs)
    freqs = np.expand_dims(freqs,0)
    if NFFT & 0x1:
        fft_res = fft_res[:,:int((NFFT+1)/2)]
        freqs = freqs[:,:int((NFFT+1)/2)]
    else:
        fft_res = fft_res[:,:int(NFFT/2)]
        freqs = freqs[:,:int(NFFT/2)]
    # fft_res = fft_res/X_col
    
    return freqs, fft_res

def gen_ref_sin(freq: float,
                srate: int,
                L: int,
                N: int,
                phase: float) -> ndarray:
    """
    Generate sine-cosine-based reference signals of one stimulus

    Parameters
    ----------
    freq : float
        Stimulus frequency
    srate : int
        Sampling rate
    L : int
        Signal length
    N : int
        Number of harmonics
    phase : float
        Stimulus phase

    Returns
    -------
    ref_sig: ndarray
        Sine-cosine-based reference signal
        2N * L
    """
    
    t = linspace(0, (L-1)/srate, L)
    t = expand_dims(t,0)
    
    y = []
    for n in range(1,N+1,1):
        y.append(sin( 2*pi*n*freq*t + n*phase ))
        y.append(cos( 2*pi*n*freq*t + n*phase ))
    y = concatenate(y, axis = 0)
        
    return y

def floor(x: float) -> int:
    """
    Floor operation. Convert np.floor result to int

    Parameters
    ----------
    x : float

    Returns
    -------
    int

    """
    
    return int(np.floor(x))


# -*- coding: utf-8 -*-
from typing import Union, Optional, Dict, List, Tuple, Callable
from numpy import ndarray

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import scipy.stats as st

def _plot_polar_phase(ax,
                      Phase : ndarray,
                      color : Optional[Union[str,list,tuple]] = None,
                      alpha : float = 0.1):
    X_shape = Phase.shape
    if len(X_shape)>1:
        Phase = np.reshape(Phase, np.prod(Phase.shape))
    for phase_val in Phase:
        ax.plot([0, phase_val], [0, 1], '-', color = color, label='_nolegend_', alpha = alpha)

def _plot_polar_phase_shadow(ax,
                             X : ndarray,
                             step_theta : float,
                             alpha = 0.3,
                             color : Optional[Union[str,list,tuple]] = None,
                             errorbar_type : str = '95ci'):
    if len(X.shape)>1:
        X = np.reshape(X, np.prod(X.shape))
    if errorbar_type.lower() == 'std':
        Y_error = np.std(X, 0)
    elif errorbar_type.lower() == '95ci':
        Y_error = cal_CI95(X)
    else:
        raise ValueError("Unknow 'errorbar_type'. 'errorbar_type' must be 'std' or '95ci'")
    Y_mean = np.mean(X)
    theta_range = np.arange(Y_mean-Y_error, Y_mean+Y_error+step_theta, step_theta)
    ax.fill_between(theta_range, 0, np.ones_like(theta_range),alpha=alpha, facecolor=color, label='_nolegend_')
    return Y_mean, Y_error
    

def polar_phase(X : ndarray,
                color: Optional[Union[str,list,tuple]] = None,
                x_label: Optional[str] = None,
                y_label: Optional[str] = None,
                x_ticks: Optional[List[str]] = None,
                legend: Optional[List[str]] = None,
                grid: bool = True,
                xlim: Optional[List[float]] = None,
                ylim: Optional[List[float]] = None,
                figsize: List[float] = [6.4, 4.8]):
    """
    Plot phase
    """
    if ylim is None:
        ylim = [0, 1.1]
    fig = plt.figure(figsize=figsize)
    ax = fig.add_axes([0,0,1,1], projection = 'polar')

    if type(X) is not list:
        if type(color) is list:
            color = color[0]
        if type(legend) is list:
            legend = legend[0]
        _plot_polar_phase(ax, X, color)
        ax.plot([0, np.mean(X)], [0, 1.1], '--', color = color, label=legend)
    else:
        if type(color) is not list:
            raise ValueError("The color must be a list.")
        if len(X) != len(color):
            raise ValueError("The length of color should be same as the length of X.")
        if type(legend) is not list:
            raise ValueError("The legend must be a list.")
        if len(X) != len(legend):
            raise ValueError("The length of legend should be same as the length of X.")
        for X_single_group, color_single_group, legend_single_group in zip(X, color, legend):
            _plot_polar_phase(ax, X_single_group, color = color_single_group)
            ax.plot([0, np.mean(X_single_group)], [0, 1.1], '--', color = color_single_group, label=legend_single_group)
    
    if x_label is not None:
        ax.set_xlabel(x_label)
    if y_label is not None:
        ax.set_ylabel(y_label)
    if x_ticks is not None:
        ax.set_xticks(X, x_ticks)
    if legend is not None:
        ax.legend()
    ax.grid(grid)
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    return fig, ax

def polar_phase_shadow(X : ndarray,
                        color: Optional[Union[str,list,tuple]] = None,
                        x_label: Optional[str] = None,
                        y_label: Optional[str] = None,
                        x_ticks: Optional[List[str]] = None,
                        legend: Optional[List[str]] = None,
                        grid: bool = True,
                        xlim: Optional[List[float]] = None,
                        ylim: Optional[List[float]] = None,
                        errorbar_type : str = '95ci',
                        alpha : float = 0.2,
                        figsize: List[float] = [6.4, 4.8]):
    """
    Plot phase with shadow
    """
    if ylim is None:
        ylim = [0, 1]
    fig = plt.figure(figsize=figsize)
    ax = fig.add_axes([0,0,1,1], projection = 'polar')
    step_theta = 0.1/180*np.pi

    if type(X) is not list:
        if type(color) is list:
            color = color[0]
        if type(legend) is list:
            legend = legend[0]
        Y_mean, _ = _plot_polar_phase_shadow(ax, X, step_theta, alpha, color, errorbar_type)
        ax.plot([0, Y_mean], [0, 1], '--', color = color, label=legend)
    else:
        if type(color) is not list:
            raise ValueError("The color must be a list.")
        if len(X) != len(color):
            raise ValueError("The length of color should be same as the length of X.")
        if type(legend) is not list:
            raise ValueError("The legend must be a list.")
        if len(X) != len(legend):
            raise ValueError("The length of legend should be same as the length of X.")
        for X_single_group, color_single_group, legend_single_group in zip(X, color, legend):
            Y_mean, _ = _plot_polar_phase_shadow(ax, X_single_group, step_theta, alpha, color_single_group, errorbar_type)
            ax.plot([0, Y_mean], [0, 1], '--', color = color_single_group, label=legend_single_group)
    
    if x_label is not None:
        ax.set_xlabel(x_label)
    if y_label is not None:
        ax.set_ylabel(y_label)
    if x_ticks is not None:
        ax.set_xticks(X, x_ticks)
    if legend is not None:
        ax.legend()
    ax.grid(grid)
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    return fig, ax

def gen_colors(color_no : int,
               colormap_name : str = 'rainbow'):
    if type(color_no) is not int:
        raise ValueError("The number of colors must be an integer.")
    color_gen = cm.get_cmap(colormap_name)(np.linspace(0,1,color_no))
    colors = []
    for i in range(color_no):
        colors.append(color_gen[i, :3])
    return colors

def close_fig(fig):
    """
    Close figure
    """
    plt.close(fig)

def _plot_hist(ax, 
                X : ndarray,
                bins: Optional[int] = None,
                range: Optional[tuple] = None,
                density: bool = True,
                color: Optional[Union[str,list,tuple]] = None,
                label: Optional[Union[str,list]] = None,
                alpha: float = 1,):
    X_shape = X.shape
    if len(X_shape)>1:
        X = np.reshape(X, np.prod(X.shape))
    if bins is None:
        bins = 'auto'
    if range is None:
        range = (X.min(), X.max())
    if color is None:
        color = 'blue'

    vals, bins, patches = ax.hist(X, bins = bins, range = range, density = density, 
                                    color = color, alpha = alpha, label = label)
    return vals, bins, patches

def _plot_fit_norm_line(ax, 
                        X : ndarray,
                        vals : ndarray,
                        range: Optional[tuple] = None,
                        line_points : int = 1000,
                        color: Optional[Union[str,list,tuple]] = None):
    X_shape = X.shape
    if len(X_shape)>1:
        X = np.reshape(X, np.prod(X.shape))
    if range is None:
        range = (X.min(), X.max())
    if color is None:
        color = 'blue'
    mu, std = st.norm.fit(X)
    x_line = np.linspace(range[0], range[1], line_points)
    y_line = st.norm.pdf(x_line, loc = mu, scale = std)
    ax.plot(x_line, y_line, '-', color = color)
    ax.plot([mu, mu], [0, np.max([y_line.max(), vals.max()])], '--', color = color)

def hist(X : Union[list, ndarray],
         bins: Optional[int] = None,
         range: Optional[tuple] = None,
         density: bool = True,
         color: Optional[Union[str,list,tuple]] = None,
         alpha: float = 1,
         fit_line: bool = True,
         line_points: int = 1000,
         x_label: Optional[str] = None,
         y_label: Optional[str] = None,
         x_ticks: Optional[List[str]] = None,
         legend: Optional[List[str]] = None,
         grid: bool = True,
         xlim: Optional[List[float]] = None,
         ylim: Optional[List[float]] = None,
         figsize: List[float] = [6.4, 4.8]):
    """
    Plot histogram
    """
    fig = plt.figure(figsize=figsize)
    ax = fig.add_axes([0,0,1,1])

    if type(X) is not list:
        if type(color) is list:
            color = color[0]
        if type(legend) is list:
            legend = legend[0]
        vals, _, _ = _plot_hist(ax, X, bins = bins, range = range, density = density, 
                                         color = color, alpha = alpha, label = legend)

        if fit_line:
            _plot_fit_norm_line(ax, X, vals, range, line_points, color)
    else:
        if type(color) is not list:
            raise ValueError("The color must be a list.")
        if len(X) != len(color):
            raise ValueError("The length of color should be same as the length of X.")
        if type(legend) is not list:
            raise ValueError("The legend must be a list.")
        if len(X) != len(legend):
            raise ValueError("The length of legend should be same as the length of X.")
        vals_list = []
        for X_single_group, color_single_group, legend_single_group in zip(X, color, legend):
            vals, _, _ = _plot_hist(ax, X_single_group, bins = bins, range = range, density = density, 
                                             color = color_single_group, alpha = alpha, label = legend_single_group)
            vals_list.append(vals)
        for X_single_group, vals, color_single_group in zip(X, vals_list, color):
            _plot_fit_norm_line(ax, X_single_group, vals, range, line_points, color_single_group)

    if x_label is not None:
        ax.set_xlabel(x_label)
    if y_label is not None:
        ax.set_ylabel(y_label)
    if x_ticks is not None:
        ax.set_xticks(X, x_ticks)
    if legend is not None:
        ax.legend()
    ax.grid(grid)
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    return fig, ax

def shadowline_plot(X: Union[list, ndarray],
                    Y: ndarray,
                    fmt: str = '-',
                    x_label: Optional[str] = None,
                    y_label: Optional[str] = None,
                    x_ticks: Optional[List[str]] = None,
                    legend: Optional[List[str]] = None,
                    errorbar_type: str = 'std',
                    grid: bool = True,
                    xlim: Optional[List[float]] = None,
                    ylim: Optional[List[float]] = None,
                    figsize: List[float] = [6.4, 4.8]):
    """
    Plot shadow lines
    Line values are equal to the mean of all observations
    Shadow areas are calculated across observations
    (x axis: variable; y axis: observation)

    Parameters
    ----------
    X : ndarray
        x value of variables
    Y : ndarray
        Plot data
        Shape: (group_num, observation_num, variable_num)
    x_label: str
        Label of x axis
        Default is None
    y_label: str
        Label of y axis
        Default is None
    x_ticks: List[str]
        Ticks of x axis
        Default is None
    legend: List[str]
        Legend of groups
        Default is None
    errorbar_type: str
        Method of calculating error, including:
            - 'std': Standard derivation
            - '95ci': 95% confidence interval
        Default is 'std'
    grid: bool
        Whether plot grid
    xlim: List[float]
        Range of x axis
    ylim: List[float]
        Range of y axis
    """
    if type(X) == ndarray:
        if len(X.shape) > 2:
            raise ValueError("Dimention of X must be smaller than 3")
    if len(Y.shape) != 3:
        raise ValueError("Plot data must have 3 dimentions")
    group_num, observation_num, variable_num = Y.shape
    if x_ticks is not None:
        if len(x_ticks) != variable_num:
            raise ValueError("Length of 'x_ticks' should be equal to 3rd dimention of data")
    if legend is not None:
        if len(legend) != group_num:
            raise ValueError("Length of 'legend' should be equal to 1st dimention of data")
            
    fig = plt.figure(figsize=figsize)
    ax = fig.add_axes([0,0,1,1])
    colors = []
    for group_idx in range(group_num):
        Y_tmp = Y[group_idx,:,:]
        Y_mean = np.mean(Y_tmp,0)
        p = ax.plot(X,Y_mean,fmt)
        colors.append(p[0].get_color())
    for group_idx in range(group_num):
        Y_tmp = Y[group_idx,:,:]
        Y_mean = np.mean(Y_tmp,0)
        if errorbar_type.lower() == 'std':
            Y_error = np.std(Y_tmp, 0)
        elif errorbar_type.lower() == '95ci':
            Y_error = cal_CI95(Y_tmp)
        else:
            raise ValueError("Unknow 'errorbar_type'. 'errorbar_type' must be 'std' or '95ci'")
        ax.fill_between(X, Y_mean-Y_error, Y_mean+Y_error ,alpha=0.3, facecolor=colors[group_idx])
    
    if x_label is not None:
        ax.set_xlabel(x_label)
    if y_label is not None:
        ax.set_ylabel(y_label)
    if x_ticks is not None:
        ax.set_xticks(X, x_ticks)
    if legend is not None:
        ax.legend(labels=legend)
    ax.grid(grid)
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    return fig, ax


def bar_plot(Y: ndarray,
             bar_sep: float = 0.25,
             x_label: Optional[str] = None,
             y_label: Optional[str] = None,
             x_ticks: Optional[List[str]] = None,
             grid: bool = True,
             xlim: Optional[List[float]] = None,
             ylim: Optional[List[float]] = None,
             figsize: List[float] = [6.4, 4.8]):
    """
    Plot bars

    For each group, a set of bars will be ploted on all variables
    Bar heights are equal to the mean of all observations
    (x axis: variable; y axis: observation)

    Parameters
    -----------
    Y: ndarray
        Plot data
        Shape: (observation_num, variable_num)
    bar_sep: Optional[float]
        Separation between two variables
        Default is 0.25
    x_label: str
        Label of x axis
        Default is None
    y_label: str
        Label of y axis
        Default is None
    x_ticks: List[str]
        Ticks of x axis
        Default is None
    grid: bool
        Whether plot grid
    xlim: List[float]
        Range of x axis
    ylim: List[float]
        Range of y axis
    """
    if len(Y.shape) > 2:
        raise ValueError("Plot data must have 3 dimentions")
    observation_num, variable_num = Y.shape
    if x_ticks is not None:
        if len(x_ticks) != variable_num:
            raise ValueError("Length of 'x_ticks' should be equal to 3rd dimention of data")

    x_center = np.arange(1, variable_num+1, 1)
    width = (1-bar_sep)/1
    
    # colors = cm.get_cmap('hsv', group_num).colors
    
    fig = plt.figure(figsize = figsize)
    ax = fig.add_axes([0,0,1,1])
    Y_mean = np.mean(Y,0)
    ax.bar(x_center, Y_mean, width = width)

    if x_label is not None:
        ax.set_xlabel(x_label)
    if y_label is not None:
        ax.set_ylabel(y_label)
    if x_ticks is not None:
        ax.set_xticks(x_center, x_ticks)
    else:
        ax.set_xticks(x_center, x_center)
    ax.grid(grid)
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    return fig, ax


def bar_plot_with_errorbar(Y: ndarray,
             bar_sep: float = 0.25,
             x_label: Optional[str] = None,
             y_label: Optional[str] = None,
             x_ticks: Optional[List[str]] = None,
             legend: Optional[List[str]] = None,
             errorbar_type: str = 'std',
             grid: bool = True,
             xlim: Optional[List[float]] = None,
             ylim: Optional[List[float]] = None,
             figsize: List[float] = [6.4, 4.8]):
    """
    Plot bars

    For each group, a set of bars will be ploted on all variables
    Bar heights are equal to the mean of all observations
    Error bar are calculated across observations
    (x axis: variable; y axis: observation)

    Parameters
    -----------
    Y: ndarray
        Plot data
        Shape: (group_num, observation_num, variable_num)
    bar_sep: Optional[float]
        Separation between two variables
        Default is 0.25
    x_label: str
        Label of x axis
        Default is None
    y_label: str
        Label of y axis
        Default is None
    x_ticks: List[str]
        Ticks of x axis
        Default is None
    legend: List[str]
        Legend of groups
        Default is None
    errorbar_type: str
        Method of calculating error, including:
            - 'std': Standard derivation
            - '95ci': 95% confidence interval
        Default is 'std'
    grid: bool
        Whether plot grid
    xlim: List[float]
        Range of x axis
    ylim: List[float]
        Range of y axis
    """
    if len(Y.shape) == 2:
        Y = np.expand_dims(Y, axis=0)
    if len(Y.shape) != 3:
        raise ValueError("Plot data must have 3 dimentions")
    group_num, observation_num, variable_num = Y.shape
    if x_ticks is not None:
        if len(x_ticks) != variable_num:
            raise ValueError("Length of 'x_ticks' should be equal to 3rd dimention of data")
    if legend is not None:
        if len(legend) != group_num:
            raise ValueError("Length of 'legend' should be equal to 1st dimention of data")

    x_center = np.arange(1, variable_num+1, 1)
    width = (1-bar_sep)/group_num
    
    # colors = cm.get_cmap('hsv', group_num).colors
    
    fig = plt.figure(figsize=figsize)
    ax = fig.add_axes([0,0,1,1])
    x = x_center - 0.5 + bar_sep/2 + width/2
    for group_idx in range(group_num):
        Y_tmp = Y[group_idx,:,:]
        Y_mean = np.mean(Y_tmp,0)
        ax.bar(x, Y_mean, width = width) #, color = colors[group_idx])
        x = x + width
    x = x_center - 0.5 + bar_sep/2 + width/2
    for group_idx in range(group_num):
        Y_tmp = Y[group_idx,:,:]
        Y_mean = np.mean(Y_tmp,0)
        if errorbar_type.lower() == 'std':
            Y_error = np.std(Y_tmp, 0)
        elif errorbar_type.lower() == '95ci':
            Y_error = cal_CI95(Y_tmp)
        else:
            raise ValueError("Unknow 'errorbar_type'. 'errorbar_type' must be 'std' or '95ci'")
        ax.errorbar(x=x, y=Y_mean, yerr=Y_error, elinewidth=2,capsize=4,fmt='none',ecolor='black')
        x = x + width

    if x_label is not None:
        ax.set_xlabel(x_label)
    if y_label is not None:
        ax.set_ylabel(y_label)
    if x_ticks is not None:
        ax.set_xticks(x_center, x_ticks)
    else:
        ax.set_xticks(x_center, x_center)
    if legend is not None:
        ax.legend(labels=legend)
    ax.grid(grid)
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    return fig, ax
        
def cal_CI95(X: ndarray) -> ndarray:
    """
    Calculate 95% confidence interval

    Parameters
    ----------
    X : ndarray

    Returns
    -------
    CI95 : ndarray
    """
    N = X.shape[0]
    SEM = np.std(X,0)/np.sqrt(N)
    CI95 = SEM * st.t.ppf(0.95, N-1)
    
    # row_num, col_num = X.shape
    # CI95 = np.zeros((2,col_num))
    # for i in range(col_num):
    #     x_tmp = X[:,i]
    #     interval = st.t.interval(0.95, df = row_num-1, loc = np.mean(x_tmp), scale=st.sem(x_tmp))
    #     CI95[0,i] = interval[0]
    #     CI95[1,i] = interval[1]
    return CI95


# -*- coding: utf-8 -*-
"""
MS-MsetCCA-R-3
"""

from typing import Union, Optional, Dict, List, Tuple, Callable
from numpy import ndarray
from joblib import Parallel, delayed
from functools import partial
from copy import deepcopy

import numpy as np

from .lsframework import TDCA_ls
from .utils_ls import (
    get_lsconfig,
    ssvep_lsframe
)
from .utils import (
    qr_list, sort, separate_trainSig, 
    blkmat
)
from .tdca import (
    _gen_delay_X, _r_tdca_canoncorr_withUV
)

def _TDCA_ls_calW(trainSig, n_jobs = None, LSconfig = None):
    def X_single_freq(trainSig_single_freq, All_train_sum):
        Nt, Np, Nc = trainSig_single_freq.shape
        X = []
        for n in range(Nt):
            X_tmp = trainSig_single_freq[n,:,:]
            X.append(X_tmp)
        X = np.concatenate(X, axis = 1)
        X = np.tile(X, (Nt,1))
        return X - 1/Nf * All_train_sum
    def Y_single_freq(trainSig_single_freq):
        Nt, Np, Nc = trainSig_single_freq.shape
        Y = []
        for n in range(Nt):
            Y_tmp = trainSig_single_freq[n,:,:] - 1/Nt * np.sum(trainSig_single_freq, axis = 0)
            Y.append(Y_tmp)
        Y = blkmat(Y)
        return Y
    #
    LSMethod, displayConvergWarn, max_n, M_error_threshold, W_error_threshold, l1_alpha, l2_alpha, alpha, reg_iter, reg_tol, check_full_rank, kernel_fun = get_lsconfig(LSconfig)
    # 
    Nf = len(trainSig)
    All_train_sum = None
    for trainSig_single_freq in trainSig:
        if All_train_sum is None:
            All_train_sum = X_single_freq(trainSig_single_freq, 0)
        else:
            All_train_sum = All_train_sum + X_single_freq(trainSig_single_freq, 0)
    if n_jobs is None:
        X = []
        for trainSig_single_freq in trainSig:
            X.append(
                X_single_freq(trainSig_single_freq, All_train_sum)
            )
            # Nt, Np, Nc = trainSig_single_freq.shape
            # for n in range(Nt):
            #     X_tmp = np.sum(trainSig_single_freq, axis = 0) - 1/Nf * All_train_sum
            #     X.append(X_tmp)
    else:
        X = Parallel(n_jobs=n_jobs)(delayed(X_single_freq)
                                                   (trainSig_single_freq = trainSig_single_freq)
                                                   for trainSig_single_freq in trainSig)
    X = np.concatenate(X, axis = 0)
    if n_jobs is None:
        Y = []
        for trainSig_single_freq in trainSig:
            Y.append(
                Y_single_freq(trainSig_single_freq)
            )
            # Nt, Np, Nc = trainSig_single_freq.shape
            # for n in range(Nt):
            #     Y_tmp = trainSig_single_freq[n,:,:] - 1/Nt * np.sum(trainSig_single_freq, axis = 0)
            #     Y.append(Y_tmp)
    else:
        Y = Parallel(n_jobs=n_jobs)(delayed(Y_single_freq)
                                            (trainSig_single_freq = trainSig_single_freq)
                                            for trainSig_single_freq in trainSig)
    Y = np.concatenate(Y, axis = 0)
    W, _, _, _, _, _ = ssvep_lsframe(X, Y,
                                    LSMethod = LSMethod,
                                    displayConvergWarn = displayConvergWarn,
                                    max_n = max_n,
                                    M_error_threshold = M_error_threshold,
                                    W_error_threshold = W_error_threshold,
                                    largest_n = max_n,
                                    check_rank = check_full_rank,
                                    l1_alpha = l1_alpha,
                                    l2_alpha = l2_alpha,
                                    alpha = alpha,
                                    reg_iter = reg_iter,
                                    reg_tol = reg_tol,
                                    kernel_fun = kernel_fun)
    W = W[-1].copy()
    return W



class MsetCCA_multi_f(TDCA_ls):
    def __init__(self,
                 n_neighbor: int = 2,
                 n_component: int = 1,
                 n_jobs: Optional[int] = None,
                 weights_filterbank: Optional[List[float]] = None,
                 n_delay: int = 0,
                 LSMethod = 'lstsq',
                 displayConvergWarn: bool = False,
                 max_n: int = 100,
                 M_error_threshold: float = 1e-6,
                 W_error_threshold: float = 1e-6,
                 l1_alpha = None,
                 l2_alpha = None,
                 alpha = None,
                 reg_iter = 1000,
                 reg_tol = 1e-4):
        super().__init__(n_component, n_jobs, weights_filterbank, n_delay, LSMethod, displayConvergWarn, max_n, M_error_threshold, W_error_threshold, l1_alpha, l2_alpha, alpha, reg_iter, reg_tol)
        self.ID = 'MsetCCA_multi_f'
        self.n_neighbor = n_neighbor

    def __copy__(self):
        copy_model = MsetCCA_multi_f(n_neighbor = self.n_neighbor,
                                     n_component = self.n_component,
                                    n_jobs = self.n_jobs,
                                    weights_filterbank = self.model['weights_filterbank'],
                                    n_delay = self.n_delay)
        copy_model.LSconfig = deepcopy(self.LSconfig)
        copy_model.model = deepcopy(self.model)
        return copy_model
    
    def fit(self,
            X: Optional[List[ndarray]] = None,
            Y: Optional[List[int]] = None,
            ref_sig: Optional[List[ndarray]] = None,
            freqs: Optional[List[float]] = None,
            *argv, **kwargv):
        if freqs is None:
            raise ValueError('ms-eTRCA requires the list of stimulus frequencies')
        if Y is None:
            raise ValueError('MsetCCA_multi_f requires training label')
        if X is None:
            raise ValueError('MsetCCA_multi_f requires training data')
        if ref_sig is None:
            raise ValueError("MsetCCA_multi_f requires reference signals")
        
        ref_sig_Q, _, _ = qr_list(ref_sig)
        ref_sig_P = [Q @ Q.T for Q in ref_sig_Q]
        self.model['ref_sig_P'] = ref_sig_P

        n_delay = self.n_delay
        X_P = _gen_delay_X(X, n_delay)
        separated_trainSig = separate_trainSig(X_P, Y)
        separated_trainSig = [np.concatenate([trainSig_single_freq, trainSig_single_freq @ P_single_freq], axis=-1)
                              for trainSig_single_freq, P_single_freq in zip(separated_trainSig, ref_sig_P)]
        
        filterbank_num, channel_num, signal_len = X_P[0].shape
        stimulus_num = len(ref_sig)
        n_component = self.n_component
        n_neighbor = self.n_neighbor
        d0 = int(np.floor(n_neighbor/2))
        _, freqs_idx, return_freqs_idx = sort(freqs)
        separated_trainSig = [separated_trainSig[i] for i in freqs_idx]

        # template_all_band = []
        trial_num = separated_trainSig[0].shape[0]
        U_tdca = np.zeros((filterbank_num, stimulus_num, channel_num, trial_num))

        for filterbank_idx in range(filterbank_num):
            for class_idx in range(1,stimulus_num+1):
                if class_idx <= d0:
                    start_idx = 0
                    end_idx = n_neighbor
                elif class_idx > d0 and class_idx < (stimulus_num-d0+1):
                    start_idx = class_idx - d0 - 1
                    end_idx = class_idx + (n_neighbor-d0-1)
                else:
                    start_idx = stimulus_num - n_neighbor
                    end_idx = stimulus_num
                separated_trainSig_tmp = [separated_trainSig[i] for i in range(start_idx, end_idx)]
                W = _TDCA_ls_calW([np.swapaxes(trainSig[:,filterbank_idx,:,:], 1,2) for trainSig in separated_trainSig_tmp],
                                self.n_jobs,
                                LSconfig = self.LSconfig)
                W_tmp = []
                for trial_idx in range(trial_num):
                    W_tmp.append(W[(trial_idx*channel_num):((trial_idx+1)*channel_num),:n_component])
                U_tdca[filterbank_idx,class_idx-1,:,:] = np.concatenate(W_tmp, axis = 1)
        
        # calculate template
        separated_trainSig = [separated_trainSig[i] for i in return_freqs_idx]
        template_sig = [np.mean(tmp, 0) for tmp in separated_trainSig]
        #     template = []
        #     for trainSig in separated_trainSig:
        #         template_tmp = []
        #         trial_num = trainSig.shape[0]
        #         for trial_idx in range(trial_num):
        #             template_single_trial = eig_vec[(trial_idx*channel_num):((trial_idx+1)*channel_num),:n_component].T @ trainSig[trial_idx,filterbank_idx,:,:]
        #             template_tmp.append(template_single_trial)
        #         template_tmp = np.concatenate(template_tmp , axis=0)
        #         template.append(template_tmp)
        #     template_all_band.append(template)

        # template_all_stimuli = []
        # for freq_i in range(stimulus_num):
        #     template_single_freq = []
        #     for filterbank_idx in range(filterbank_num):
        #         template_single_freq.append(np.expand_dims(template_all_band[filterbank_idx][freq_i], axis = 0))
        #     template_single_freq = np.concatenate(template_single_freq, axis = 0)
        #     template_all_stimuli.append(template_single_freq)
        
        # self.model['template'] = template_all_stimuli
        # template_sig_Q, template_sig_R, template_sig_P = qr_list(template_all_stimuli)
        # self.model['template_sig_Q'] = template_sig_Q # List of shape: (stimulus_num,);
        # self.model['template_sig_R'] = template_sig_R
        # self.model['template_sig_P'] = template_sig_P

        # U_tdca = np.repeat(U_tdca, repeats = stimulus_num, axis = 1)
        self.model['U'] = U_tdca[:,return_freqs_idx,:,:]
        self.model['template_sig'] = template_sig

    def predict(self,
                X: List[ndarray]) -> List[int]:
        weights_filterbank = self.model['weights_filterbank']
        if weights_filterbank is None:
            weights_filterbank = [1 for _ in range(X[0].shape[0])]
        if type(weights_filterbank) is list:
            weights_filterbank = np.expand_dims(np.array(weights_filterbank),1).T
        else:
            if len(weights_filterbank.shape) != 2:
                raise ValueError("'weights_filterbank' has wrong shape")
            if weights_filterbank.shape[0] != 1:
                weights_filterbank = weights_filterbank.T
        if weights_filterbank.shape[0] != 1:
            raise ValueError("'weights_filterbank' has wrong shape")
        n_delay = self.n_delay

        X_delay = _gen_delay_X(X, n_delay)

        template_sig = self.model['template_sig']
        U = self.model['U'] 
        ref_sig_P = self.model['ref_sig_P']

        if self.n_jobs is not None:
            r = Parallel(n_jobs=self.n_jobs)(delayed(partial(_r_tdca_canoncorr_withUV, Y=template_sig, P=ref_sig_P, U=U, V=U))(X=a) for a in X_delay)
        else:
            r = []
            for a in X_delay:
                r.append(
                    _r_tdca_canoncorr_withUV(X=a, Y=template_sig, P=ref_sig_P, U=U, V=U)
                )

        Y_pred = [int( np.argmax( weights_filterbank @ r_tmp)) for r_tmp in r]
        
        return Y_pred, r
    
# def _r_cca_qr_withP(X: ndarray,
#            Y_Q: List[ndarray],
#            Y_R: List[ndarray],
#            Y_P: List[ndarray],
#            P: List[ndarray],
#            n_component: int,
#            force_output_UV: Optional[bool] = False) -> Union[ndarray, Tuple[ndarray, ndarray, ndarray]]:
#     """
#     Calculate correlation of CCA based on QR decomposition for single trial data 

#     Parameters
#     ----------
#     X : ndarray
#         Single trial EEG data
#         EEG shape: (filterbank_num, channel_num, signal_len)
#     Y_Q : List[ndarray]
#         Q of reference signals
#     Y_R: List[ndarray]
#         R of reference signals
#     Y_P: List[ndarray]
#         P of reference signals
#     n_component : int
#         Number of eigvectors for spatial filters.
#     force_output_UV : Optional[bool]
#         Whether return spatial filter 'U' and weights of harmonics 'V'

#     Returns
#     -------
#     R : ndarray
#         Correlation
#         shape: (filterbank_num * stimulus_num)
#     U : ndarray
#         Spatial filter
#         shape: (filterbank_num * stimulus_num * channel_num * n_component)
#     V : ndarray
#         Weights of harmonics
#         shape: (filterbank_num * stimulus_num * harmonic_num * n_component)
#     """
#     filterbank_num, channel_num, signal_len = X.shape
#     harmonic_num = Y_R[0].shape[-1]
#     stimulus_num = len(Y_Q)
    
#     Y = [qr_inverse(Y_Q[i],Y_R[i],Y_P[i]) for i in range(len(Y_Q))]
#     if len(Y[0].shape)==2: # reference
#         Y = [Y_tmp.T for Y_tmp in Y]
#     elif len(Y[0].shape)==3: # template
#         Y = [np.transpose(Y_tmp, (0,2,1)) for Y_tmp in Y]
#     else:
#         raise ValueError('Unknown data type')
    
#     # R1 = np.zeros((filterbank_num,stimulus_num))
#     # R2 = np.zeros((filterbank_num,stimulus_num))
#     R = np.zeros((filterbank_num, stimulus_num))
#     U = np.zeros((filterbank_num, stimulus_num, channel_num, n_component))
#     V = np.zeros((filterbank_num, stimulus_num, harmonic_num, n_component))
    
#     for k in range(filterbank_num):
#         tmp_X = X[k,:,:]
#         for i in range(stimulus_num):
#             tmp = np.concatenate([tmp_X, tmp_X @ P[i]], axis = -1)
#             X_Q, X_R, X_P = qr_remove_mean(tmp.T)
#             if len(Y_Q[i].shape)==2: # reference
#                 Y_Q_tmp = Y_Q[i]
#                 Y_R_tmp = Y_R[i]
#                 Y_P_tmp = Y_P[i]
#                 Y_tmp = Y[i]
#             elif len(Y_Q[i].shape)==3: # template
#                 Y_Q_tmp = Y_Q[i][k,:,:]
#                 Y_R_tmp = Y_R[i][k,:,:]
#                 Y_P_tmp = Y_P[i][k,:]
#                 Y_tmp = Y[i][k,:,:]
#             else:
#                 raise ValueError('Unknown data type')
#             svd_X = X_Q.T @ Y_Q_tmp
#             if svd_X.shape[0]>svd_X.shape[1]:
#                 full_matrices=False
#             else:
#                 full_matrices=True
            
#             if n_component == 0 and force_output_UV is False:
#                 D = svd(svd_X, full_matrices, False)
#                 r = D[0]
#             else:
#                 L, D, M = svd(svd_X, full_matrices, True)
#                 M = M.T
#                 try:
#                     A = mldivide(X_R, L) * np.sqrt(signal_len - 1)
#                     B = mldivide(Y_R_tmp, M) * np.sqrt(signal_len - 1)
#                 except:
#                     print('Error')
#                 A_r = np.zeros(A.shape)
#                 for n in range(A.shape[0]):
#                     A_r[X_P[n],:] = A[n,:]
#                 B_r = np.zeros(B.shape)
#                 for n in range(B.shape[0]):
#                     B_r[Y_P_tmp[n],:] = B[n,:]
                
#                 a = A_r[:channel_num, :n_component].T @ tmp
#                 b = B_r[:harmonic_num, :n_component].T @ Y_tmp
#                 a = np.reshape(a, (-1))
#                 b = np.reshape(b, (-1))
                
#                 # r2 = stats.pearsonr(a, b)[0]
#                 # r = stats.pearsonr(a, b)[0]
#                 r = np.corrcoef(a, b)[0,1]
#                 U[k,i,:,:] = A_r[:channel_num, :n_component]
#                 V[k,i,:,:] = B_r[:harmonic_num, :n_component]
                
#             # R1[k,i] = r1
#             # R2[k,i] = r2
#             R[k,i] = r
#     if force_output_UV:
#         return R, U, V
#     else:
#         return R

        



class eMsetCCA_multi_f(TDCA_ls):
    def __init__(self,
                 n_neighbor: int = 2,
                 n_component: int = 1,
                 n_jobs: Optional[int] = None,
                 weights_filterbank: Optional[List[float]] = None,
                 n_delay: int = 0,
                 LSMethod = 'lstsq',
                 displayConvergWarn: bool = False,
                 max_n: int = 100,
                 M_error_threshold: float = 1e-6,
                 W_error_threshold: float = 1e-6,
                 l1_alpha = None,
                 l2_alpha = None,
                 alpha = None,
                 reg_iter = 1000,
                 reg_tol = 1e-4):
        super().__init__(n_component, n_jobs, weights_filterbank, n_delay, LSMethod, displayConvergWarn, max_n, M_error_threshold, W_error_threshold, l1_alpha, l2_alpha, alpha, reg_iter, reg_tol)
        self.ID = 'eMsetCCA_multi_f'
        self.n_neighbor = n_neighbor

    def __copy__(self):
        copy_model = eMsetCCA_multi_f(n_neighbor = self.n_neighbor,
                                     n_component = self.n_component,
                                    n_jobs = self.n_jobs,
                                    weights_filterbank = self.model['weights_filterbank'],
                                    n_delay = self.n_delay)
        copy_model.LSconfig = deepcopy(self.LSconfig)
        copy_model.model = deepcopy(self.model)
        return copy_model
    
    def fit(self,
            X: Optional[List[ndarray]] = None,
            Y: Optional[List[int]] = None,
            ref_sig: Optional[List[ndarray]] = None,
            freqs: Optional[List[float]] = None,
            *argv, **kwargv):
        if freqs is None:
            raise ValueError('ms-eTRCA requires the list of stimulus frequencies')
        if Y is None:
            raise ValueError('MsetCCA_multi_f requires training label')
        if X is None:
            raise ValueError('MsetCCA_multi_f requires training data')
        if ref_sig is None:
            raise ValueError("MsetCCA_multi_f requires reference signals")
        
        ref_sig_Q, _, _ = qr_list(ref_sig)
        ref_sig_P = [Q @ Q.T for Q in ref_sig_Q]
        self.model['ref_sig_P'] = ref_sig_P

        n_delay = self.n_delay
        X_P = _gen_delay_X(X, n_delay)
        separated_trainSig = separate_trainSig(X_P, Y)
        separated_trainSig = [np.concatenate([trainSig_single_freq, trainSig_single_freq @ P_single_freq], axis=-1)
                              for trainSig_single_freq, P_single_freq in zip(separated_trainSig, ref_sig_P)]
        
        filterbank_num, channel_num, signal_len = X_P[0].shape
        stimulus_num = len(ref_sig)
        n_component = self.n_component
        n_neighbor = min(self.n_neighbor,stimulus_num)
        d0 = int(np.floor(n_neighbor/2))
        _, freqs_idx, return_freqs_idx = sort(freqs)
        separated_trainSig = [separated_trainSig[i] for i in freqs_idx]

        # template_all_band = []
        trial_num = separated_trainSig[0].shape[0]
        U_tdca = np.zeros((filterbank_num, 1, channel_num, trial_num*stimulus_num))

        for filterbank_idx in range(filterbank_num):
            W_tmp = []
            for class_idx in range(1,stimulus_num+1):
                if class_idx <= d0:
                    start_idx = 0
                    end_idx = n_neighbor
                elif class_idx > d0 and class_idx < (stimulus_num-d0+1):
                    start_idx = class_idx - d0 - 1
                    end_idx = class_idx + (n_neighbor-d0-1)
                else:
                    start_idx = stimulus_num - n_neighbor
                    end_idx = stimulus_num
                separated_trainSig_tmp = [separated_trainSig[i] for i in range(start_idx, end_idx)]
                W = _TDCA_ls_calW([np.swapaxes(trainSig[:,filterbank_idx,:,:], 1,2) for trainSig in separated_trainSig_tmp],
                                self.n_jobs,
                                LSconfig = self.LSconfig)
                for trial_idx in range(trial_num):
                    W_tmp.append(W[(trial_idx*channel_num):((trial_idx+1)*channel_num),:n_component])
            U_tdca[filterbank_idx,0,:,:] = np.concatenate(W_tmp, axis = 1)
        
        # calculate template
        separated_trainSig = [separated_trainSig[i] for i in return_freqs_idx]
        template_sig = [np.mean(tmp, 0) for tmp in separated_trainSig]
        #     template = []
        #     for trainSig in separated_trainSig:
        #         template_tmp = []
        #         trial_num = trainSig.shape[0]
        #         for trial_idx in range(trial_num):
        #             template_single_trial = eig_vec[(trial_idx*channel_num):((trial_idx+1)*channel_num),:n_component].T @ trainSig[trial_idx,filterbank_idx,:,:]
        #             template_tmp.append(template_single_trial)
        #         template_tmp = np.concatenate(template_tmp , axis=0)
        #         template.append(template_tmp)
        #     template_all_band.append(template)

        # template_all_stimuli = []
        # for freq_i in range(stimulus_num):
        #     template_single_freq = []
        #     for filterbank_idx in range(filterbank_num):
        #         template_single_freq.append(np.expand_dims(template_all_band[filterbank_idx][freq_i], axis = 0))
        #     template_single_freq = np.concatenate(template_single_freq, axis = 0)
        #     template_all_stimuli.append(template_single_freq)
        
        # self.model['template'] = template_all_stimuli
        # template_sig_Q, template_sig_R, template_sig_P = qr_list(template_all_stimuli)
        # self.model['template_sig_Q'] = template_sig_Q # List of shape: (stimulus_num,);
        # self.model['template_sig_R'] = template_sig_R
        # self.model['template_sig_P'] = template_sig_P

        U_tdca = np.repeat(U_tdca, repeats = stimulus_num, axis = 1)
        self.model['U'] = U_tdca
        self.model['template_sig'] = template_sig

    def predict(self,
                X: List[ndarray]) -> List[int]:
        weights_filterbank = self.model['weights_filterbank']
        if weights_filterbank is None:
            weights_filterbank = [1 for _ in range(X[0].shape[0])]
        if type(weights_filterbank) is list:
            weights_filterbank = np.expand_dims(np.array(weights_filterbank),1).T
        else:
            if len(weights_filterbank.shape) != 2:
                raise ValueError("'weights_filterbank' has wrong shape")
            if weights_filterbank.shape[0] != 1:
                weights_filterbank = weights_filterbank.T
        if weights_filterbank.shape[0] != 1:
            raise ValueError("'weights_filterbank' has wrong shape")
        n_delay = self.n_delay

        X_delay = _gen_delay_X(X, n_delay)

        template_sig = self.model['template_sig']
        U = self.model['U'] 
        ref_sig_P = self.model['ref_sig_P']

        if self.n_jobs is not None:
            r = Parallel(n_jobs=self.n_jobs)(delayed(partial(_r_tdca_canoncorr_withUV, Y=template_sig, P=ref_sig_P, U=U, V=U))(X=a) for a in X_delay)
        else:
            r = []
            for a in X_delay:
                r.append(
                    _r_tdca_canoncorr_withUV(X=a, Y=template_sig, P=ref_sig_P, U=U, V=U)
                )

        Y_pred = [int( np.argmax( weights_filterbank @ r_tmp)) for r_tmp in r]
        
        return Y_pred, r



# -*- coding: utf-8 -*-

import os
import numpy as np
import py7zr

from typing import Union, Optional, Dict, List, Tuple
from numpy import ndarray, transpose

from .basedataset import BaseDataset
from .subjectinfo import SubInfo
from ..utils.download import download_single_file
from ..utils.io import loadmat

class BenchmarkDataset(BaseDataset):
    """
    Benchmark Dataset
    
    This dataset gathered SSVEP-BCI recordings of 35 healthy subjects (17 females, aged 17-34 years, mean age: 22 years) focusing on 40 characters flickering at different frequencies (8-15.8 Hz with an interval of 0.2 Hz).
    For each subject, the experiment consisted of 6 blocks. Each block contained 40 trials corresponding to all 40 characters indicated in a random order. Each trial started with a visual cue (a red square) indicating a target stimulus. The cue appeared for 0.5 s on the screen.
    Following the cue offset, all stimuli started to flicker on the screen concurrently and lasted 5 s.
    After stimulus offset, the screen was blank for 0.5 s before the next trial began, which allowed the subjects to have short breaks between consecutive trials.
    Each trial lasted a total of 6 s.

    Total: around 3.45 GB
    
    Paper:
    Y. Wang, X. Chen, X. Gao, and S. Gao, “A benchmark dataset for SSVEP-based braincomputer interfaces,” IEEE Trans. Neural Syst. Rehabil. Eng., vol. 25, no. 10, pp. 17461752, 2017.
    """
 
    _CHANNELS = [
        'FP1','FPZ','FP2','AF3','AF4','F7','F5','F3','F1','FZ','F2',
        'F4','F6','F8','FT7','FC5','FC3','FC1','FCz','FC2','FC4','FC6',
        'FT8','T7','C5','C3','C1','Cz','C2','C4','C6','T8','M1','TP7',
        'CP5','CP3','CP1','CPZ','CP2','CP4','CP6','TP8','M2','P7','P5',
        'P3','P1','PZ','P2','P4','P6','P8','PO7','PO5','PO3','POz','PO4',
        'PO6','PO8','CB1','O1','Oz','O2','CB2'
    ]

    _FREQS = [
        8, 9, 10, 11, 12, 13, 14, 15, 
        8.2, 9.2, 10.2, 11.2, 12.2, 13.2, 14.2, 15.2, 
        8.4, 9.4, 10.4, 11.4, 12.4, 13.4, 14.4, 15.4,
        8.6, 9.6, 10.6, 11.6, 12.6, 13.6, 14.6, 15.6,
        8.8, 9.8, 10.8, 11.8, 12.8, 13.8, 14.8, 15.8
    ]

    _PHASES = [
        0, 0.5, 1, 1.5, 0, 0.5, 1, 1.5,
        0.5, 1, 1.5, 0, 0.5, 1, 1.5, 0,
        1, 1.5, 0, 0.5, 1, 1.5, 0, 0.5,
        1.5, 0, 0.5, 1, 1.5, 0, 0.5, 1,
        0, 0.5, 1, 1.5, 0, 0.5, 1, 1.5
    ]
    
    _SUBJECTS = [SubInfo(ID = 'S{:d}'.format(sub_idx)) for sub_idx in range(1,35+1,1)]
    
    def __init__(self, 
                 path: Optional[str] = None,
                 path_support_file: Optional[str] = None):
        super().__init__(subjects = self._SUBJECTS, 
                         ID = 'Benchmark Dataset', 
                         url = 'http://bci.med.tsinghua.edu.cn/upload/yijun/', 
                         paths = path, 
                         channels = self._CHANNELS, 
                         srate = 250, 
                         block_num = 6, 
                         trial_num = len(self._FREQS),
                         trial_len = 6, 
                         stim_info = {'stim_num': len(self._FREQS),
                                      'freqs': self._FREQS,
                                      'phases': [i * np.pi for i in self._PHASES]},
                         support_files = ['Readme.txt',
                                          'Sub_info.txt',
                                          '64-channels.loc',
                                          'Freq_Phase.mat'],
                         path_support_file = path_support_file,
                         t_prestim = 0.5,
                         t_break = 0.5,
                         default_t_latency = 0.14)
        self.trial_label_check_list = {}
        for trial_i in range(self.trial_num):
            self.trial_label_check_list[trial_i] = trial_i

    
    def download_single_subject(self,
                                subject: SubInfo):
        source_url = self.url + subject.ID + '.mat.7z'
        desertation = os.path.join(subject.path, subject.ID + '.mat.7z')
        
        data_file = os.path.join(subject.path, subject.ID + '.mat')

        download_flag = True
        
        if not os.path.isfile(data_file):
            try:
                download_single_file(source_url, desertation)
            
                with py7zr.SevenZipFile(desertation,'r') as archive:
                    archive.extractall(subject.path)
                    
                os.remove(desertation)
            except:
                download_flag = False
        
        return download_flag, source_url, desertation
    
    def download_file(self,
                      file_name: str):
        source_url = self.url + file_name
        desertation = os.path.join(self.path_support_file, file_name)

        download_flag = True
        
        if not os.path.isfile(desertation):
            try:
                download_single_file(source_url, desertation)
            except:
                download_flag = False
        
        return download_flag, source_url, desertation
        
    def get_sub_data(self, 
                     sub_idx: int) -> ndarray:
        if sub_idx < 0:
            raise ValueError('Subject index cannot be negative')
        if sub_idx > len(self.subjects)-1:
            raise ValueError('Subject index should be smaller than {:d}'.format(len(self.subjects)))
        
        sub_info = self.subjects[sub_idx]
        file_path = os.path.join(sub_info.path, sub_info.ID + '.mat')
        
        mat_data = loadmat(file_path)
        data = mat_data['data']
        data = transpose(data, (3,2,0,1)) # block_num * stimulus_num * ch_num * whole_trial_samples
        
        return data
    
    def get_label_single_trial(self,
                               sub_idx: int,
                               block_idx: int,
                               trial_idx: int) -> int:
        return self.trial_label_check_list[trial_idx]
        


# -*- coding: utf-8 -*-
"""
Base class of evaluator
"""

import abc
from typing import Union, Optional, Dict, List, Tuple, Callable
from numpy import ndarray

from tqdm import tqdm
import time
import numpy as np
import os
import pickle
import copy

from itertools import combinations

from joblib import Parallel, delayed
from functools import partial

import warnings

def gen_trials_onedataset_cross_subj_notarget_diffsignlen_specfic_trainblcokNum(dataset_idx: int,
                                         tw_seq: List[float],
                                         dataset_container: list,
                                         harmonic_num: int,
                                         # trials: List[int],
                                         ch_used: List[int],
                                         # trainblockNum: int,
                                         train_subjects: Optional[List[int]],
                                         test_subjects: Optional[List[int]],
                                         t_latency: Optional[float] = None,
                                         shuffle: bool = False) -> list:
    """
    Generate cross-subject evaluation trials for one dataset
    Evaluations will be carried out on each signal length for given pair of train and test subjects

    Parameters
    ----------
    dataset_idx : int
        dataset index of dataset_container
    tw_seq : List[float]
        List of signal length
    dataset_container : list
        List of datasets
    harmonic_num : int
        Number of harmonics
    ch_used : List[int]
        List of channels
    train_subjects : Optional[List[int]]
        List of subject indices in training dataset
    test_subjects : Optional[List[int]]
        List of subject indices in testing dataset
    t_latency : Optional[float]
        Latency time
        If None, default latency time of dataset will be used
    shuffle : bool
        Whether shuffle

    Returns
    -------
    trial_container : list
        List of trial information

    """
    AllBlockInd = [i for i in range(dataset_container[dataset_idx].block_num)]
    all_trials = [i for i in range(dataset_container[dataset_idx].trial_num)]
    
    trial_container = []
    for tw in tw_seq:
        train_trial = TrialInfo().add_dataset(dataset_idx = dataset_idx,
                                                sub_idx = train_subjects,
                                                block_idx = AllBlockInd,
                                                trial_idx = all_trials,
                                                ch_idx = ch_used,
                                                harmonic_num = harmonic_num,
                                                tw = tw,
                                                t_latency = t_latency,
                                                shuffle = shuffle)
        test_trial = TrialInfo().add_dataset(dataset_idx = dataset_idx,
                                                sub_idx = test_subjects,
                                                block_idx = AllBlockInd,
                                                trial_idx = all_trials,
                                                ch_idx = ch_used,
                                                harmonic_num = harmonic_num,
                                                tw = tw,
                                                t_latency = t_latency,
                                                shuffle = shuffle)
        trial_container.append([train_trial, test_trial])
    return trial_container
        

def gen_trials_onedataset_individual_online(dataset_idx: int,
                                         tw_seq: List[float],
                                         dataset_container: list,
                                         harmonic_num: int,
                                         repeat_num: int,
                                         trials: List[int],
                                         ch_used: List[int],
                                         t_latency: Optional[float] = None,
                                         shuffle: bool = False) -> list:
    """
    Generate evaluation trials for one dataset
    Evaluations will be carried out on each subject and each signal length
    Training datasets are the 1st block
    Testing datasets are the all blocks

    Parameters
    ----------
    dataset_idx : int
        dataset index of dataset_container
    tw : List[float]
        signal length
    sub_idx : int
        Subject index
    dataset_container : list
        List of datasets
    harmonic_num : int
        Number of harmonics
    repeat_num : int
        Number of randon times
    trials: List[int]
        List of trial index
    ch_used : List[int]
        List of channels
    t_latency : Optional[float]
        Latency time
        If None, default latency time of dataset will be used
    shuffle : bool
        Whether shuffle

    Returns
    -------
    trial_container : list
        List of trial information

    """
    sub_num = len(dataset_container[dataset_idx].subjects)
    trial_container = []
    for tw in tw_seq:
        for sub_idx in range(sub_num):
            for _ in range(repeat_num):
                train_block = [0]
                test_block = [block_idx for block_idx in range(dataset_container[dataset_idx].block_num)]
                train_trial = TrialInfo().add_dataset(dataset_idx = dataset_idx,
                                                        sub_idx = sub_idx,
                                                        block_idx = train_block,
                                                        trial_idx = trials,
                                                        ch_idx = ch_used,
                                                        harmonic_num = harmonic_num,
                                                        tw = tw,
                                                        t_latency = t_latency,
                                                        shuffle = shuffle)
                test_trial = TrialInfo().add_dataset(dataset_idx = dataset_idx,
                                                        sub_idx = sub_idx,
                                                        block_idx = test_block,
                                                        trial_idx = trials,
                                                        ch_idx = ch_used,
                                                        harmonic_num = harmonic_num,
                                                        tw = tw,
                                                        t_latency = t_latency,
                                                        shuffle = shuffle)
                trial_container.append([train_trial, test_trial])
    return trial_container

def gen_trials_onedataset_individual_diffsignlen_specfic_trainblcokNum(dataset_idx: int,
                                         tw_seq: List[float],
                                         dataset_container: list,
                                         harmonic_num: int,
                                         trials: List[int],
                                         ch_used: List[int],
                                         trainblockNum: int,
                                         subjects: Optional[List[int]] = None,
                                         t_latency: Optional[float] = None,
                                         shuffle: bool = False) -> list:
    """
    Generate evaluation trials for one dataset with specific training block number
    Evaluations will be carried out on each subject and each signal length

    Parameters
    ----------
    dataset_idx : int
        dataset index of dataset_container
    tw_seq : List[float]
        List of signal length
    dataset_container : list
        List of datasets
    harmonic_num : int
        Number of harmonics
    trials: List[int]
        List of trial index
    ch_used : List[int]
        List of channels
    trainblockNum : int
        Number of training blocks
    subjects : Optional[List[int]]
        List of subject indices
        If None, all subjects will be included
    t_latency : Optional[float]
        Latency time
        If None, default latency time of dataset will be used
    shuffle : bool
        Whether shuffle

    Returns
    -------
    trial_container : list
        List of trial information

    """
    if subjects is None:
        sub_num = len(dataset_container[dataset_idx].subjects)
        subjects = list(range(sub_num))
    AllBlockInd = [i for i in range(dataset_container[dataset_idx].block_num)]
    train_block_list = []
    for c in combinations(AllBlockInd, trainblockNum):
        train_block_list.append(list(c))
    trial_container = []
    for tw in tw_seq:
        for sub_idx in subjects:
            # for block_idx in range(dataset_container[dataset_idx].block_num):
            for train_block in train_block_list:
                # test_block, train_block = dataset_container[dataset_idx].leave_one_block_out(block_idx)
                test_block = AllBlockInd.copy()
                for train_block_ind in train_block:
                    test_block.remove(train_block_ind)
                train_trial = TrialInfo().add_dataset(dataset_idx = dataset_idx,
                                                      sub_idx = sub_idx,
                                                      block_idx = train_block,
                                                      trial_idx = trials,
                                                      ch_idx = ch_used,
                                                      harmonic_num = harmonic_num,
                                                      tw = tw,
                                                      t_latency = t_latency,
                                                      shuffle = shuffle)
                test_trial = TrialInfo().add_dataset(dataset_idx = dataset_idx,
                                                      sub_idx = sub_idx,
                                                      block_idx = test_block,
                                                      trial_idx = trials,
                                                      ch_idx = ch_used,
                                                      harmonic_num = harmonic_num,
                                                      tw = tw,
                                                      t_latency = t_latency,
                                                      shuffle = shuffle)
                trial_container.append([train_trial, test_trial])
    return trial_container

def gen_trials_onedataset_individual_diffsiglen(dataset_idx: int,
                                         tw_seq: List[float],
                                         dataset_container: list,
                                         harmonic_num: int,
                                         trials: List[int],
                                         ch_used: List[int],
                                         subjects: Optional[List[int]] = None,
                                         t_latency: Optional[float] = None,
                                         shuffle: bool = False) -> list:
    """
    Generate evaluation trials for one dataset
    Evaluations will be carried out on each subject and each signal length
    Training and testing datasets are separated based on the leave-one-block-out rule

    Parameters
    ----------
    dataset_idx : int
        dataset index of dataset_container
    tw_seq : List[float]
        List of signal length
    dataset_container : list
        List of datasets
    harmonic_num : int
        Number of harmonics
    trials: List[int]
        List of trial index
    ch_used : List[int]
        List of channels
    subjects : Optional[List[int]]
        List of subject indices
        If None, all subjects will be included
    t_latency : Optional[float]
        Latency time
        If None, default latency time of dataset will be used
    shuffle : bool
        Whether shuffle

    Returns
    -------
    trial_container : list
        List of trial information

    """
    if subjects is None:
        sub_num = len(dataset_container[dataset_idx].subjects)
        subjects = list(range(sub_num))
    trial_container = []
    for tw in tw_seq:
        for sub_idx in subjects:
            for block_idx in range(dataset_container[dataset_idx].block_num):
                test_block, train_block = dataset_container[dataset_idx].leave_one_block_out(block_idx)
                train_trial = TrialInfo().add_dataset(dataset_idx = dataset_idx,
                                                      sub_idx = sub_idx,
                                                      block_idx = train_block,
                                                      trial_idx = trials,
                                                      ch_idx = ch_used,
                                                      harmonic_num = harmonic_num,
                                                      tw = tw,
                                                      t_latency = t_latency,
                                                      shuffle = shuffle)
                test_trial = TrialInfo().add_dataset(dataset_idx = dataset_idx,
                                                      sub_idx = sub_idx,
                                                      block_idx = test_block,
                                                      trial_idx = trials,
                                                      ch_idx = ch_used,
                                                      harmonic_num = harmonic_num,
                                                      tw = tw,
                                                      t_latency = t_latency,
                                                      shuffle = shuffle)
                trial_container.append([train_trial, test_trial])
    return trial_container

class pbarParallel(Parallel):
    def __init__(self, 
                 loop_list_num: List[int],
                 use_tqdm : bool = True,
                 desc: str = '',
                 *args, **kwargs):
        self.loop_list_num = loop_list_num
        self.use_tqdm = use_tqdm
        self.desc = desc
        super().__init__(*args, **kwargs)
    def __call__(self, *args, **kwargs):
        with create_pbar(use_tqdm = self.use_tqdm, loop_list_num = self.loop_list_num, desc = self.desc) as self._pbar:
            return Parallel.__call__(self, *args, **kwargs)
    def print_progress(self):
        # self._pbar.total = self.n_dispatched_tasks
        self._pbar.n = self.n_completed_tasks
        self._pbar.refresh()

def create_pbar(loop_list_num: List[int],
                use_tqdm : bool = True,
                desc: str = ''):
    """
    Create process bar

    Parameters
    ----------
    loop_list_num : List[int]
        List of following loop numbers
    desc : str
        Prefix for the progressbar

    Returns
    -------
    pbar : tqdm
        processbar
    """
    total_num = 1
    for loop_num in loop_list_num:
        total_num = total_num * loop_num
    pbar = tqdm(disable=not use_tqdm,
                total = total_num,
                desc = desc,
                bar_format = '{desc}{percentage:3.3f}%|{bar}| {n_fmt}/{total_fmt} [Time: {elapsed}<{remaining}]',
                dynamic_ncols = True)
    return pbar

class PerformanceContainer:
    """
    Container of performance, containing:
        - true_label_train: True labels for train
        - pred_label_train: Predicted labels for train
        - true_label_test: True labels for test
        - pred_label_test: Preducted labels for test
        - train_time: Training times
        - test_time_train: Testing times for train
        - test_time_test: Testing times for test
    """
    def __init__(self, 
                 method_ID: str):
        self.method_ID = method_ID
        self.clear()
        
    def add_true_label_train(self,true_label:list):
        self.true_label_train.extend(true_label)
    def add_pred_label_train(self,pred_label:list):
        self.pred_label_train.extend(pred_label)
    def add_pred_r_train(self,pred_r:list):
        self.pred_r_train.extend(pred_r)
    def add_true_label_test(self,true_label:list):
        self.true_label_test.extend(true_label)
    def add_pred_label_test(self,pred_label:list):
        self.pred_label_test.extend(pred_label)
    def add_pred_r_test(self,pred_r:list):
        self.pred_r_test.extend(pred_r)
    def add_train_time(self,train_time:list):
        self.train_time.append(train_time)
    def add_test_time_train(self,test_time:list):
        self.test_time_train.append(test_time)
    def add_test_time_test(self,test_time:list):
        self.test_time_test.append(test_time)
    def clear(self):
        self.true_label_train = []
        self.pred_label_train = []
        self.pred_r_train = []
        self.true_label_test = []
        self.pred_label_test = []
        self.pred_r_test = []
        self.train_time = []
        self.test_time_train = []
        self.test_time_test = []

class TrialInfo:
    """
    TrialInfo
    
    Store training and testing trial information
    
    If there are more than two datasets, sine-cone-based reference signals will 
    be generated based on the first dataset information
    """
    def __init__(self):
        """
        Parameters
        ----------
        dataset_idx : list
            List of dataset index
        sub_idx : list
            List of subjects
            Length is equal to dataset_idx
            Each element is a list of subjext index for one dataset
        block_idx : list
            List of blocks
            Length is equal to dataset_idx
            Each element is a list of block index for one dataset
        trial_idx : list
            List of trials
            Length is equal to dataset_idx
            Each element is a list of trial index for one dataset
        ch_idx : list
            List of channels
            Length is equal to dataset_idx
            Each element is a list of channel index for one dataset
        t_latency ; list
            List of latency time
            Length  is equal to dataset_idx
            Each element is a list of latency time for one dataset
        harmonic_num : int
            Number of harmonics for sine-cosine-based references
        tw : float
            Signal window length
        shuffle : List[bool]
            Whether shuffle trials
            Length is equal to dataset_idx
            Each element is the number of shuffle flag for one dataset
        """
        
        self.dataset_idx = []
        self.sub_idx = []
        self.block_idx = []
        self.trial_idx = []
        self.ch_idx = []
        self.harmonic_num = []
        self.tw = []
        self.t_latency = []
        self.shuffle = []
                
    def add_dataset(self,
                    dataset_idx : int,
                    sub_idx : List[int],
                    block_idx : List[int],
                    trial_idx : List[int],
                    ch_idx : List[int],
                    harmonic_num : List[int],
                    tw : float,
                    t_latency : Optional[float] = None,
                    shuffle : bool = False):
        self.dataset_idx.append(dataset_idx)
        self.sub_idx.append(sub_idx)
        self.block_idx.append(block_idx)
        self.trial_idx.append(trial_idx)
        self.ch_idx.append(ch_idx)
        self.t_latency.append(t_latency)
        self.harmonic_num = harmonic_num
        self.tw = tw
        self.shuffle.append(shuffle)
        return self
    
    def get_data(self,
                 dataset_container: list,
                 ignore_stim_phase: bool) -> Tuple[list, list, list, list]:
        """
        Get data based trian information

        Parameters
        ----------
        dataset_container : list
            List of datasets

        Returns
        -------
        X: list
            Data
        Y: list
            Labels
        ref_sig: list
            Reference signal
        freqs: list
            List of stimulus frequencies corresponding to reference signal
        """
        dataset = dataset_container[self.dataset_idx[0]]
        
        trial_ind = self.trial_idx[0].copy()
        trial_ind.sort()

        ref_sig = dataset.get_ref_sig(self.tw,self.harmonic_num, ignore_stim_phase)
        ref_sig = [ref_sig[i] for i in trial_ind]
        freqs = dataset.stim_info['freqs']
        freqs = [freqs[i] for i in trial_ind]

        X = []
        Y = []
        for (dataset_idx,
            sub_idx,
            block_idx,
            trial_idx,
            ch_idx,
            t_latency,
            shuffle) in zip(self.dataset_idx, 
                           self.sub_idx,
                           self.block_idx,
                           self.trial_idx,
                           self.ch_idx,
                           self.t_latency,
                           self.shuffle):
            dataset = dataset_container[dataset_idx]
            X_tmp, Y_tmp = dataset.get_data(sub_idx = sub_idx,
                                            blocks = block_idx,
                                            trials = trial_idx,
                                            channels = ch_idx,
                                            sig_len = self.tw,
                                            t_latency = t_latency,
                                            shuffle = shuffle)
            X.extend(X_tmp)
            Y.extend(Y_tmp)
        return X, Y, ref_sig, freqs
        

def _run_loop(trial_idx, trial_container, 
              model_container, dataset_container, ignore_stim_phase,
              eval_train,
              save_model):
    trial = trial_container[trial_idx]

    # Create performance for one trial
    performance_one_trial = [PerformanceContainer(model.ID) for model in model_container]
    
    # Get train data
    train_trial_info = trial[0]
    # if self.disp_processbar:
    #     print('-------train info------------')
    #     print(train_trial_info.__dict__)
    if len(train_trial_info.dataset_idx) == 0:
        raise ValueError('Train trial {:d} information is empty'.format(trial_idx))
    X, Y, ref_sig, freqs = train_trial_info.get_data(dataset_container, ignore_stim_phase)
    
    # Train models 
    model_one_trial = []
    for train_model_idx, model_tmp in enumerate(model_container):
        trained_model = model_tmp.__copy__()
        tic = time.time()
        trained_model.fit(X=X, Y=Y, ref_sig=ref_sig, freqs=freqs) 
        performance_one_trial[train_model_idx].add_train_time(time.time()-tic)
        model_one_trial.append(trained_model)
        # if disp_processbar:
        #     pbar.update(pbar_update_val)
    if eval_train:
        for test_model_idx, model_tmp in enumerate(model_one_trial):
            tic = time.time()
            pred_label, pred_r = model_tmp.predict(X)
            performance_one_trial[test_model_idx].add_test_time_train(time.time()-tic)
            performance_one_trial[test_model_idx].add_pred_label_train(pred_label)
            performance_one_trial[test_model_idx].add_pred_r_train(pred_r)
            performance_one_trial[test_model_idx].add_true_label_train(Y)
            # if disp_processbar:
            #     pbar.update(pbar_update_val)

    # Get test data
    test_trial_info = trial[1]
    # if self.disp_processbar:
    #     print('-------test info------------')
    #     print(test_trial_info.__dict__)
    if len(test_trial_info.dataset_idx) == 0:
        raise ValueError('Test trial {:d} information is empty'.format(trial_idx))
    X, Y, ref_sig, _ = test_trial_info.get_data(dataset_container, ignore_stim_phase)
        
    # Test models
    for test_model_idx, model_tmp in enumerate(model_one_trial):
        tic = time.time()
        pred_label, pred_r = model_tmp.predict(X)
        # print(np.array(Y)-np.array(pred_label))
        performance_one_trial[test_model_idx].add_test_time_test(time.time()-tic)
        performance_one_trial[test_model_idx].add_pred_label_test(pred_label)
        performance_one_trial[test_model_idx].add_pred_r_test(pred_r)
        performance_one_trial[test_model_idx].add_true_label_test(Y)
        # if disp_processbar:
        #     pbar.update(pbar_update_val)

    if save_model:
        return performance_one_trial, model_one_trial
    else:
        return performance_one_trial, None
        
    # self.performance_container.append(performance_one_trial)
    # if self.save_model:
    #     self.trained_model_container.append(model_one_trial)

class BaseEvaluator:
    """
    BaseEvaluator
    """
    def __init__(self,
                 dataset_container : Optional[list] = None,
                 model_container : Optional[list] = None,
                 trial_container : Optional[list] = None,
                 save_model: bool = False,
                 disp_processbar: bool = True,
                 ignore_stim_phase: bool = False):
        """
        Parameters for all evaluators

        Parameters
        ----------
        dataset_container : list
            List of datasets
        model_container : list
            List of models
        trial_container : list
            List of trians
            Each element is a list of two elements that are training and testing trials
        save_model : Optional[bool], optional
            Whether save models
        disp_processbar : Optional[bool], optional
            Whether display process bar
        ignore_stim_phase: bool
            Whether ignore stimulus phases when generating reference signals
        """
        self.dataset_container = dataset_container
        self.model_container = model_container
        self.trial_container = trial_container
        self.save_model = save_model
        self.disp_processbar = disp_processbar
        self.ignore_stim_phase = ignore_stim_phase
        
        self.performance_container = []
        self.trained_model_container = []

    def save(self,
             file: str):
        desertation_dir = os.path.dirname(file)
        if not os.path.exists(desertation_dir):
            os.makedirs(desertation_dir)
        if os.path.isfile(file):
            os.remove(file)
        with open(file,'wb') as file_:
            try:
                pickle.dump(self, file_, pickle.HIGHEST_PROTOCOL)
            except:
                warnings.warn("Cannot save whole evaluator. So remove 'dataset_container' and try to save it again.")
                saved_self = copy.deepcopy(self)
                try:
                    saved_self.dataset_container = None
                    pickle.dump(saved_self, file_, pickle.HIGHEST_PROTOCOL)
                except:
                    warnings.warn("Remove 'dataset_container' still cannot be saved. So remove 'model_container' and 'trained_model_container' and try to save it again.")
                    try:
                        saved_self.model_container = None
                        saved_self.trained_model_container = None
                        pickle.dump(saved_self, file_, pickle.HIGHEST_PROTOCOL)
                    except:
                        warnings.warn("Remove 'model_container' and 'trained_model_container' still cannot be saved. So only save 'performance_container'.")
                        pickle.dump(self.performance_container, file_, pickle.HIGHEST_PROTOCOL)

    def load(self,
             file: str,
             only_performance: bool = False):
        if not os.path.isfile(file):
            raise ValueError('{:s} does not exist!!'.format(file))
        with open(file,'rb') as file_:
            self_load = pickle.load(file_)
        if type(self_load) is not list and type(self_load) is not tuple:
            if only_performance:
                self.performance_container = copy.deepcopy(self_load.performance_container)
                self.trial_container = copy.deepcopy(self_load.trial_container)
            else:
                self.dataset_container = copy.deepcopy(self_load.dataset_container)
                self.model_container = copy.deepcopy(self_load.model_container)
                self.trial_container = copy.deepcopy(self_load.trial_container)
                self.save_model = copy.deepcopy(self_load.save_model)
                self.disp_processbar = copy.deepcopy(self_load.disp_processbar)
                self.ignore_stim_phase = copy.deepcopy(self_load.ignore_stim_phase)
                self.performance_container = copy.deepcopy(self_load.performance_container)
                self.trained_model_container = copy.deepcopy(self_load.trained_model_container)
        else:
            self.performance_container = copy.deepcopy(self_load)
            warnings.warn("'{:s}' only contains a list or a tuple. So only 'performance_container' is reloaded.".format(file))
        
    def run(self,
            n_jobs : Optional[int] = None,
            timeout : Optional[int] = None,
            eval_train : bool = False):
        """
        Run evaluator

        Parameters
        ----------
        n_jobs : Optional[int], optional
            Number of CPUs. The default is None.
        eval_train : bool
            Whether evaluate train performance
        save_model_after_evaluate : bool
            Whether save the model after the evaluation. To evaluate the online method, this flag is set as True
        """
        if self.dataset_container is None or self.model_container is None or self.trial_container is None:
            raise ValueError("Please check 'dataset_container', 'model_container', and 'trial_container'.")
        # if n_jobs is not None:
        #     for i in range(len(self.model_container)):
        #         self.model_container[i].n_jobs = n_jobs
        for i in range(len(self.model_container)):
            self.model_container[i].n_jobs = None
                
        if self.disp_processbar:
            print('\n========================\n   Start\n========================\n')
            # pbar = create_pbar([len(self.trial_container)])
            # if eval_train:
            #     pbar = create_pbar([len(self.trial_container), len(self.model_container)*3])
            # else:
            #     pbar = create_pbar([len(self.trial_container), len(self.model_container)*2])
            # pbar_update_val = 1
                
        self.performance_container, self.trained_model_container = zip(*pbarParallel(n_jobs=n_jobs,timeout=timeout,loop_list_num=[len(self.trial_container)],use_tqdm=self.disp_processbar)
                                                                                               (delayed(partial(_run_loop, model_container = self.model_container,
                                                                                                                           trial_container = self.trial_container,
                                                                                                                           dataset_container = self.dataset_container,
                                                                                                                           ignore_stim_phase = self.ignore_stim_phase,
                                                                                                                           eval_train = eval_train,
                                                                                                                           save_model = self.save_model
                                                                                                                           ))(trial_idx = trial_idx) 
                                                                                                                           for trial_idx in range(len(self.trial_container))))
        
        if self.disp_processbar:
            # pbar.close()
            print('\n========================\n   End\n========================\n')     
            
    def search_trial_idx(self,
                         train_or_test: str,
                         trial_info: dict) -> list:
        """
        According to given trial information, search trials

        Parameters
        ----------
        train_or_test : str
            Given trial information related to training or testing trials
        trial_info : dict
            Trial information. It can contain following items:
                - dataset_idx : list
                    List of dataset index
                sub_idx : list
                    List of subjects
                block_idx : list
                    List of blocks
                trial_idx : list
                    List of trials
                ch_idx : list
                    List of channels
                t_latency ; list
                    List of latency time
                harmonic_num : int
                    Number of harmonics for sine-cosine-based references
                tw : float
                    Signal window length
                shuffle : List[bool]
                    Whether shuffle trials

        Returns
        -------
        trial_idx: list
            List of indices of trials that satify the given information

        """
        if self.dataset_container is None or self.model_container is None or self.trial_container is None:
            raise ValueError("Please check 'dataset_container', 'model_container', and 'trial_container'.")
        if train_or_test.lower() == "train":
            idx = 0
        elif train_or_test.lower() == "test":
            idx = 1
        else:
            raise ValueError("Unknown train_or_test type. It must be 'train' or 'test'")
        res = []
        for trial_idx, trial in enumerate(self.trial_container):
            store_flag = True
            for search_key, search_value in trial_info.items():
                store_flag = store_flag and (trial[idx].__dict__[search_key] == search_value)
            if store_flag:
                res.append(trial_idx)
                
        return res
            

        
                


# -*- coding: utf-8 -*-

from numpy import ndarray
from typing import Union, Optional, Dict, List, Tuple
from scipy import signal

import numpy as np

def suggested_weights_filterbank(num_subbands: Optional[int] = 5) -> List[float]:
    """
    Provide suggested weights of filterbank for benchmark dataset

    Returns
    -------
    weights_filterbank : List[float]
        Suggested weights of filterbank
    """
    return [i**(-1.25)+0.25 for i in range(1,num_subbands+1,1)]

def suggested_ch() -> List[int]:
    """
    Provide suggested channels for benchmark dataset

    Returns
    -------
    ch_used: List
        Suggested channels (PZ, PO5, PO3, POz, PO4, PO6, O1, Oz, O2)
    """
    return [i-1 for i in [48, 54, 55, 56, 57, 58, 61, 62, 63]]

def occipital_19_ch() -> List[int]:
    """
    Provide 19 channels around occipital region for benchmark dataset

    Returns
    -------
    ch_used: List
        19 channels in occipital regions (P7, P5, P3, P1, PZ, P2, P4, P6,
        P8, PO7, PO5, PO3, POz, PO4, PO6, PO8, O1, Oz, O2)
    """
    return [i-1 for i in [44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 61, 62, 63]]

def center_occipital_26_ch() -> List[int]:
    """
    Provide 26 channels from center region to occipital region for benchmark dataset

    Returns
    -------
    ch_used: List
        26 channels from center region to occipital region 
        (CP5, CP3, CP1, CPZ, CP2, CP4, CP6,
         P7, P5, P3, P1, PZ, P2, P4, P6, P8,
         PO7, PO5, PO3, POz, PO4, PO6, PO8
         O1, Oz, O2)
    """
    return [i-1 for i in [35, 36, 37, 38, 39, 40, 41,
                          44, 45, 46, 47, 48, 49, 50, 51, 52,
                          53, 54, 55, 56, 57, 58, 59,
                          61, 62, 63]]

def center_occipital_33_ch() -> List[int]:
    """
    Provide 33 channels from center region to occipital region for benchmark dataset

    Returns
    -------
    ch_used: List
        33 channels from center region to occipital region 
        (C5, C3, C1, Cz, C2, C4, C6,
         CP5, CP3, CP1, CPZ, CP2, CP4, CP6,
         P7, P5, P3, P1, PZ, P2, P4, P6, P8,
         PO7, PO5, PO3, POz, PO4, PO6, PO8
         O1, Oz, O2)
    """
    return [i-1 for i in [25, 26, 27, 28, 29, 30, 31,
                          35, 36, 37, 38, 39, 40, 41,
                          44, 45, 46, 47, 48, 49, 50, 51, 52,
                          53, 54, 55, 56, 57, 58, 59,
                          61, 62, 63]]

def preprocess(dataself,
               X: ndarray) -> ndarray:
    """
    Suggested preprocessing function for benchmark dataset
    
    notch filter at 50 Hz
    """
    srate = dataself.srate

    # notch filter at 50 Hz
    f0 = 50
    Q = 35
    notchB, notchA = signal.iircomb(f0, Q, ftype='notch', fs=srate)
    preprocess_X = signal.filtfilt(notchB, notchA, X, axis = 1, padtype='odd', padlen=3*(max(len(notchB),len(notchA))-1))
    
    return preprocess_X
    
def filterbank(dataself,
               X: ndarray,
               num_subbands: Optional[int] = 5) -> ndarray:
    """
    Suggested filterbank function for benchmark dataset
    """
    srate = dataself.srate
    
    filterbank_X = np.zeros((num_subbands, X.shape[0], X.shape[1]))
    
    for k in range(1, num_subbands+1, 1):
        Wp = [(8*k)/(srate/2), 90/(srate/2)]
        Ws = [(8*k-2)/(srate/2), 100/(srate/2)]

        gstop = 40
        while gstop>=20:
            try:
                N, Wn = signal.cheb1ord(Wp, Ws, 3, gstop)
                bpB, bpA = signal.cheby1(N, 0.5, Wn, btype = 'bandpass')
                filterbank_X[k-1,:,:] = signal.filtfilt(bpB, bpA, X, axis = 1, padtype='odd', padlen=3*(max(len(bpB),len(bpA))-1))
                break
            except:
                gstop -= 1
        if gstop<20:
            raise ValueError("""
Filterbank cannot be processed. You may try longer signal lengths.
Filterbank order: {:n}
gstop: {:n}
bpB: {:s}
bpA: {:s}
Required signal length: {:n}
Signal length: {:n}""".format(k,
                                gstop,
                                str(bpB),
                                str(bpA),
                                3*(max(len(bpB),len(bpA))-1),
                                X.shape[1]))
        
        
    return filterbank_X


# -*- coding: utf-8 -*-

from numpy import ndarray, object
from typing import Union, Optional, Dict, List, Tuple
import warnings

import scipy.io as sio
import mat73
import numpy as np
import os
import pickle

def savedata(file: str,
             data: dict,
             save_type: str = 'mat'):
    """
    Save data

    Parameters
    ----------
    file : str
        File name 
    data : dict
        Saving data
        Keys are variable names
        Values are variable values
    save_type : str, optional
        File type
        'mat' - matlab file in .mat format
        'np' - Binary file in NumPy .npy format
        The default is 'mat'.
    """
    if file == '':
        raise ValueError('The input "file" cannot be empty.')

    desertation_dir = os.path.dirname(file)
    if desertation_dir == '':
        pass
    elif not os.path.exists(desertation_dir):
        os.makedirs(desertation_dir)
    
    if save_type.lower() == 'mat':
        if os.path.isfile(file):
            os.remove(file)
        try:
            sio.savemat(file, data)
        except:
            warnings.warn("Cannot save data as 'mat' file!! So the data is saved as a numpy data file.")
            file = os.path.splitext(file)[0]+'.npy'
            savedata(file, data, 'np')
    elif save_type.lower() == 'np':
        if os.path.isfile(file):
            os.remove(file)
        try:
            with open(file, 'wb') as f:
                np.save(f, data, allow_pickle=True, fix_imports=True)
        except:
            warnings.warn("Cannot save data as 'np' file!! So the data is saved as a binary data file.")
            file = os.path.splitext(file)[0]+'.pkl'
            with open(file, 'wb') as f:
                pickle.dump(data, f, protocol = 4) # Load file: pickle.load(file)
    else:
        raise ValueError("Unknown 'save_type'. 'save_type' only can be 'mat' or 'np'")
        
def loaddata(file: str,
             save_type: str = 'mat') -> dict:
    """
    Load data

    Parameters
    ----------
    file : str
        File name
    save_type : str, optional
        File type
        'mat' - matlab file in .mat format
        'np' - Binary file in NumPy .npy format
        The default is 'mat'.

    Return
    -----------
    data: dict
        Loaded data
    """
    if save_type.lower() == 'mat':
        data = loadmat(file)
        if not os.path.splitext(file)[1].lower() == '.mat':
            warnings.warn("The following file seems not MATLAB file (.mat): {:s}".format(file))
    elif save_type.lower() == 'np':
        with open(file, 'rb') as f:
            data = np.load(f, allow_pickle=True)
            data = data.item()
        if not os.path.splitext(file)[1].lower() == '.npy':
            warnings.warn("The following file seems not numpy file (.npy): {:s}".format(file))
    else:
        raise ValueError("Unknown 'save_type'. 'save_type' only can be 'mat' or 'np'")
        
    return data

def loadmat(file_path: str) -> dict:
    """
    Load mat file

    Parameters
    -------------------
    file_path: str
        Full file path

    Returns
    -------
    mat_data: dict
        Data in mat file
    """

    try:
        data = _loadmat(file_path)
    except:
        data = mat73.loadmat(file_path)

    return data

def _loadmat(filename):
    '''
    this function should be called instead of direct sio.loadmat
    as it cures the problem of not properly recovering python dictionaries
    from mat files. It calls the function check keys to cure all entries
    which are still mat-objects
    Notes: only works for mat before matlab v7.3
    '''
    def _check_keys(d):
        '''
        checks if entries in dictionary are mat-objects. If yes
        todict is called to change them to nested dictionaries
        '''
        for key in d:
            if isinstance(d[key], sio.matlab.mio5_params.mat_struct):
                d[key] = _todict(d[key])
            elif isinstance(d[key], ndarray):
                d[key] = _tolist(d[key])
        return d

    def _todict(matobj):
        '''
        A recursive function which constructs from matobjects nested dictionaries
        '''
        d = {}
        for strg in matobj._fieldnames:
            elem = matobj.__dict__[strg]
            if isinstance(elem, sio.matlab.mio5_params.mat_struct):
                d[strg] = _todict(elem)
            elif isinstance(elem, ndarray):
                d[strg] = _tolist(elem)
            else:
                d[strg] = elem
        return d

    def _tolist(elem):
        '''
        A recursive function which constructs lists from cellarrays
        (which are loaded as numpy ndarrays), recursing into the elements
        if they contain matobjects.
        '''
        if elem.dtype == object:
            elem_list = []
            for sub_elem in elem:
                if isinstance(sub_elem, sio.matlab.mio5_params.mat_struct):
                    elem_list.append(_todict(sub_elem))
                elif isinstance(sub_elem, ndarray):
                    elem_list.append(_tolist(sub_elem))
                else:
                    elem_list.append(sub_elem)
            return elem_list
        else:
            return elem

    data = sio.loadmat(filename, struct_as_record=False, squeeze_me=True)
    return _check_keys(data)


# -*- coding: utf-8 -*-

from .baseevaluator import (
    BaseEvaluator, 
    gen_trials_onedataset_individual_diffsignlen_specfic_trainblcokNum,
    gen_trials_onedataset_individual_diffsiglen,
    gen_trials_onedataset_individual_online,
    gen_trials_onedataset_cross_subj_notarget_diffsignlen_specfic_trainblcokNum
)

from .performance import (
    cal_confusionmatrix_onedataset_individual_online,
    cal_performance_onedataset_individual_online,
    cal_confusionmatrix_onedataset_individual_diffsiglen,
    cal_performance_onedataset_individual_diffsiglen,
    cal_itr_trials_onebyone, cal_itr_trials, 
    cal_acc_trials_onebyone, cal_acc_trials,
    cal_acc, cal_itr,
    cal_performance_onedataset_crossSubj_diffsiglen
)

from .plot import (
    close_fig,
    shadowline_plot, bar_plot, bar_plot_with_errorbar, cal_CI95,
    hist, gen_colors,
    polar_phase, polar_phase_shadow
)


# -*- coding: utf-8 -*-

from numpy import ndarray
from typing import Union, Optional, Dict, List, Tuple
from scipy import signal

import numpy as np

def subj_idx_highperformance(subj_no: int = 49):
    subj_idx_from_matlab = [4,99,84,87,69,79,24,39,91,7,54,28,66,30,38,98,72,53,19,11,92,65,3,5,29,62,50,23,83,16,18,52,9,17,25,56,46,78,82,61,81,80,33,40,93,12,75,100,20,26,55,86,70,51,74,89,90,88,41,68,97,36,49,27,31,8,42,95,22,76,37,71,44,59,47,14,6,96,58,77,85,60,94,45,67,64,43,21,57,2,102,48,35,63,13,1,73,34,15,32,10,101]
    subj_idx_for_python = [i-1 for i in subj_idx_from_matlab]
    return subj_idx_for_python[0:subj_no]

def suggested_weights_filterbank(num_subbands: int = 5,
                                 data_type: str = 'wet',
                                 method_type:str = 'cca'):
    """
    Provide weights of filterbank
    """
    if method_type.lower() == 'cca':
        if data_type.lower() == 'wet':
            return [i**(-1.25)+0 for i in range(1,num_subbands+1,1)]
        else:
            return [i**(-2)+0.25 for i in range(1,num_subbands+1,1)]
    else:
        if data_type.lower() == 'wet':
            return [i**(-1.75)+0.5 for i in range(1,num_subbands+1,1)]
        else:
            return [i**(-1.25)+0.25 for i in range(1,num_subbands+1,1)]
    # return [i**(-1.25)+0.25 for i in range(1,num_subbands+1,1)]

def suggested_ch() -> List[int]:
    """
    Provide suggested channels

    Returns
    -------
    ch_used: List
        Suggested channels
    """
    return list(range(8))

def preprocess(dataself,
               X: ndarray) -> ndarray:
    """
    Suggested preprocessing function
    
    notch filter at 50 Hz
    """
    srate = dataself.srate

    # notch filter at 50 Hz
    # f0 = 50
    # Q = 35
    # notchB, notchA = signal.iircomb(f0, Q, ftype='notch', fs=srate)
    # preprocess_X = signal.filtfilt(notchB, notchA, X, axis = 1, padtype='odd', padlen=3*(max(len(notchB),len(notchA))-1))

    b1, a1 = signal.cheby1(4,2,[47.0/(srate/2), 53.0/(srate/2)], 'bandstop')
    preprocess_X = signal.filtfilt(b1, a1, X, axis = 1, padtype='odd', padlen=3*(max(len(b1),len(a1))-1))
    
    return preprocess_X
    
def filterbank(dataself,
               X: ndarray,
               num_subbands: Optional[int] = 5) -> ndarray:
    """
    Suggested filterbank function
    """
    srate = dataself.srate
    
    filterbank_X = np.zeros((num_subbands, X.shape[0], X.shape[1]))
    
    for k in range(1, num_subbands+1, 1):
        b2, a2 = signal.cheby1(4,1,[(9.0*k)/(srate/2), 90.0/(srate/2)],'bandpass')
        tmp =  signal.filtfilt(b2, a2, X, axis = 1, padtype='odd', padlen=3*(max(len(b2),len(a2))-1))
        tmp = signal.detrend(tmp, axis = 1)
        tmp = tmp - np.mean(tmp, axis = 1, keepdims = True)
        tmp_std = np.std(tmp, axis = 1, keepdims = True, ddof=1)
        tmp = tmp / tmp_std
        filterbank_X[k-1,:,:] = tmp.copy()
        
#         Wp = [(9.25*k)/(srate/2), 90/(srate/2)]
#         Ws = [(9.25*k-2)/(srate/2), 100/(srate/2)]

#         gstop = 40
#         while gstop>=20:
#             try:
#                 N, Wn = signal.cheb1ord(Wp, Ws, 3, gstop)
#                 bpB, bpA = signal.cheby1(N, 0.5, Wn, btype = 'bandpass')
#                 filterbank_X[k-1,:,:] = signal.filtfilt(bpB, bpA, X, axis = 1, padtype='odd', padlen=3*(max(len(bpB),len(bpA))-1))
#                 break
#             except:
#                 gstop -= 1
#         if gstop<20:
#             raise ValueError("""
# Filterbank cannot be processed. You may try longer signal lengths.
# Filterbank order: {:n}
# gstop: {:n}
# bpB: {:s}
# bpA: {:s}
# Required signal length: {:n}
# Signal length: {:n}""".format(k,
#                                 gstop,
#                                 str(bpB),
#                                 str(bpA),
#                                 3*(max(len(bpB),len(bpA))-1),
#                                 X.shape[1]))
        
        
    return filterbank_X


# -*- coding: utf-8 -*-
"""
Base class of SSVEP recognition method
"""

import abc
from typing import Union, Optional, Dict, List, Tuple, Callable
from numpy import ndarray

class BaseModel(metaclass=abc.ABCMeta):
    """
    BaseModel
    """
    def __init__(self,
                 ID: str,
                 n_component: int = 1,
                 n_jobs: Optional[int] = None,
                 weights_filterbank: Optional[List[float]] = None):
        """
        Parameters for all SSVEP recognition methods

        Parameters
        ----------
        ID : str
            Unique identifier for method
        n_component : Optional[int], optional
            Number of eigvectors for spatial filters. The default is 1.
        n_jobs : Optional[int], optional
            Number of CPU for computing different trials. The default is None.
        weights_filterbank : Optional[List[float]], optional
            Weights of spatial filters. 
            If None, all weights are 1.
            The default is None.
        """
        if n_component < 0:
            raise ValueError('n_component must be larger than 0')
        
        self.ID = ID
        self.n_component = n_component
        self.n_jobs = n_jobs
        
        self.model = {}
        self.model['weights_filterbank'] = weights_filterbank
        
    @abc.abstractclassmethod
    def fit(self, *argv, **kwargs):
        """
        Training function

        Different methods may require different parameters.

        Spatial filters and other parameters will be stored in self.model
        """
        pass
    
    @abc.abstractclassmethod
    def predict(self,
                X: List[ndarray]) -> List[int]:
        """
        Testing function

        Parameters
        ----------
        X : List[ndarray]
            List of testing EEG data
            List shape: (trial_num,)
            EEG shape: (filterbank_num, channel_num, signal_len)

        Returns
        -------
        Y : List[int]
            List of recognition labels (stimulus indices)
            List shape: (trial_num,)
        """
        pass
    
    @abc.abstractclassmethod
    def __copy__(self):
        """
        copy itself
        """
        pass